/*
 * Copyright (c) 1999 by Sun Microsystems, Inc.
 * All rights reserved.
 */

/* 
 * Portions may be based on code from the linux ext2 filesystem 
 * 
 *  Copyright (C) 1993  Stephen Tweedie (sct@dcs.ed.ac.uk)
 *  Copyright (C) 1992  Remy Card (card@masi.ibp.fr)
 *                      Laboratoire MASI - Institut Blaise Pascal
 *                      Universite Pierre et Marie Curie (Paris VI)
 * Copyright (C) 1991, 1992  Linus Torvalds
 */

#pragma ident	"%Z%%M%	%I%	%E% SMI"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kmem.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/cred.h>
#include <sys/disp.h>
#include <sys/buf.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/fdio.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/conf.h>
#undef NFSCLIENT
#include <sys/statvfs.h>
#include <sys/mount.h>
#include <sys/pathname.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/sysmacros.h>
#include <sys/conf.h>
#include <sys/mkdev.h>
#include <sys/swap.h>
#include <sys/sunddi.h>
#include <sys/dktp/fdisk.h>
#include <fs/fs_subr.h>
#include <sys/modctl.h>
#include <sys/vol.h>
#include <sys/dkio.h>
#include <sys/open.h>
#include <sys/flock.h>
#include <sys/dirent.h>
#include <sys/stat.h>
#include <vm/seg_vn.h>
#include <sys/vmsystm.h>
#include <vm/pvn.h>
#include <ext2fs.h>


static int ext2fs_open(struct vnode **vpp, int flag, struct cred *cr);
static int ext2fs_close(struct vnode *vp, int flag, int count, offset_t offset,
    struct cred *cr);
static int ext2fs_read(struct vnode *vp, struct uio *uiop, int ioflag,
    struct cred *cr);
static int ext2fs_write(struct vnode *vp, struct uio *uiop, int ioflag,
    struct cred *cr);
static int ext2fs_getattr(struct vnode *vp, struct vattr *vap, int flags,
    struct cred *cr);
static int ext2fs_setattr(struct vnode *vp, struct vattr *vap, int flags,
    struct cred *cr);
static int ext2fs_access(struct vnode *vp, int mode, int flags, struct cred *cr);
static int ext2fs_lookup(struct vnode *dvp, char *nm, struct vnode **vpp,
    struct pathname *pnp, int flags, struct vnode *rdir, struct cred *cr);
static int ext2fs_create(struct vnode *dvp, char *name, struct vattr *vap,
    enum vcexcl excl, int mode, struct vnode **vpp, struct cred *cr, int flag);
static int ext2fs_remove(struct vnode *vp, char *nm, struct cred *cr);
static int ext2fs_link(struct vnode *tdvp, struct vnode *svp, char *tnm,
    struct cred *cr);
static int ext2fs_rename(struct vnode *sdvp, char *snm, struct vnode *tdvp,
    char *tnm, struct cred *cr);
static int ext2fs_mkdir(struct vnode *dvp, char *dirname, struct vattr *vap,
    struct vnode **vpp, struct cred *cr);
static int ext2fs_rmdir(struct vnode *vp, char *nm, struct vnode *cdir,
    struct cred *cr);
static int ext2fs_readdir(struct vnode *vp, struct uio *uiop, struct cred *cr,
    int *eofp);
static int ext2fs_symlink(struct vnode *dvp, char *linkname, struct vattr *vap,
    char *target, struct cred *cr);
static int ext2fs_readlink(struct vnode *vp, struct uio *uiop, struct cred *cr);
static int ext2fs_fsync(struct vnode *vp, int syncflag, struct cred *cr);
static void ext2fs_inactive(struct vnode *vp, struct cred *cr);
static int ext2fs_fid(struct vnode *vp, struct fid *fidp);
static void ext2fs_rwlock(struct vnode *vp, int write_lock);
static void ext2fs_rwunlock(struct vnode *vp, int write_lock);
static int ext2fs_seek(struct vnode *vp, offset_t ooff, offset_t *noffp);
static int ext2fs_frlock(struct vnode *vp, int cmd, struct flock64 *bfp,
    int flag, offset_t offset, cred_t *cr);
static int ext2fs_space(struct vnode *vp, int cmd, struct flock64 *bfp, int flag,
    offset_t offset, struct cred *cr);
static int ext2fs_getpage(struct vnode *vp, offset_t off, size_t len,
    uint_t *protp, page_t *plarr[], size_t plsz, struct seg *seg, caddr_t addr,
    enum seg_rw rw, struct cred *cr);
static int ext2fs_putapage(struct vnode *vp, page_t *pp, u_offset_t *offp,
    size_t *lenp, int flags, struct cred *cr);
static int ext2fs_putpage(struct vnode *vp, offset_t off, size_t len, int flags,
    struct cred *cr);
static int ext2fs_map(struct vnode *vp, offset_t off, struct as *as,
    caddr_t *addrp, size_t len, uchar_t prot, uchar_t maxprot, uint_t flags,
    struct cred *cr);
static int ext2fs_addmap(struct vnode *vp, offset_t off, struct as *as,
    caddr_t addr, size_t len, uchar_t prot, uchar_t maxprot, uint_t flags,
    struct cred *cr);
static int ext2fs_delmap(struct vnode *vp, offset_t off, struct as *as,
    caddr_t addr, size_t len, uint_t prot, uint_t maxprot, uint_t flags,
    struct cred *cr);
static int ext2fs_l_pathconf(struct vnode *vp, int cmd, ulong_t *valp,
    struct cred *cr);
static int ext2fs_dumpctl(vnode_t *vp, int free);
static int ext2fs_getsecattr(struct vnode *vp, vsecattr_t *vsap, int flag,
    struct cred *cr);
static int ext2fs_setsecattr(struct vnode *vp, vsecattr_t *vsap, int flag,
    struct cred *cr);
static int ext2fs_dump(vnode_t *vp, caddr_t addr, int ldbn, int dblks);

struct vnodeops ext2fs_vnodeops = {
	ext2fs_open,	/* will not be blocked by lockfs */
	ext2fs_close,	/* will not be blocked by lockfs */
	ext2fs_read,
	ext2fs_write,
	fs_nosys,
	fs_setfl,
	ext2fs_getattr,
	ext2fs_setattr,
	ext2fs_access,
	ext2fs_lookup,
	ext2fs_create,
	ext2fs_remove,
	ext2fs_link,
	ext2fs_rename,
	ext2fs_mkdir,
	ext2fs_rmdir,
	ext2fs_readdir,
	ext2fs_symlink,
	ext2fs_readlink,
	ext2fs_fsync,
	ext2fs_inactive,	/* will not be blocked by lockfs */
	ext2fs_fid,
	ext2fs_rwlock,		/* will not be blocked by lockfs */
	ext2fs_rwunlock,	/* will not be blocked by lockfs */
	ext2fs_seek,
	fs_cmp,
	ext2fs_frlock,
	ext2fs_space,
	fs_nosys,	/* realvp */
	ext2fs_getpage,
	ext2fs_putpage,
	ext2fs_map,
	ext2fs_addmap,	/* will not be blocked by lockfs */
	ext2fs_delmap,	/* will not be blocked by lockfs */
	fs_poll,
	ext2fs_dump,	/* dump */
	ext2fs_l_pathconf,
	fs_nosys,	/* pageio */
	ext2fs_dumpctl,
	fs_dispose,
	fs_nosys,	/* setsecattr */
	fs_fab_acl,	/* getsecattr */
	fs_shrlock	/* shrlock */
};

static int
ext2fs_open(struct vnode **vpp, int flag, struct cred *cr)
{
	return (0);
}

static int
ext2fs_close(struct vnode *vp, int flag, int count, offset_t offset,
    struct cred *cr)
{
	cleanlocks(vp, ttoproc(curthread)->p_pid, 0);
	cleanshares(vp, ttoproc(curthread)->p_pid);
	return (0);
}

static int
ext2fs_read(struct vnode *vp, struct uio *uiop, int ioflag, struct cred *cr)
{
	int	error = 0;
	struct inode *ip = VTOI(vp);
	mode_t	type;
	u_int	blkno;
	u_int	offset;
	struct buf *bp;
	u_int	amount;
	u_int	bufremain;
	u_int	bufoff;
	struct ext2fs *fsp;
	caddr_t	holebuf;
	caddr_t	bufaddr;

	type = ip->i_mode & IFMT;
	if ((type != IFREG) && (type != IFDIR) && (type != IFLNK))
		return (EIO);
	if (uiop->uio_loffset > MAXOFFSET_T)
		return (0);
	if (uiop->uio_resid == 0)
		return (0);

	if (uiop->uio_loffset >= ip->i_size)
		return (0);

	fsp = ip->i_fs;
	/*
	mutex_enter(&fsp->e_lock);
	 */
	rw_enter(&ip->i_rwlock, RW_READER);
	offset = uiop->uio_loffset;
	blkno = offset >> EXT2_BLOCK_SIZE_BITS(fsp);
	bufoff = offset - blkno * EXT2_BLOCK_SIZE(fsp);
	bufremain = EXT2_BLOCK_SIZE(fsp) - bufoff;
	holebuf = NULL;
nextblk:
	bp = ext2fs_bread(ip, blkno, 0, &error);
	if (bp == NULL) {
		if (error == EFBIG) {
			/* hole */
			holebuf = kmem_zalloc(EXT2_BLOCK_SIZE(fsp), KM_SLEEP);
			bufaddr = holebuf;
		} else {
			goto rwexitout;
		}
	} else {
		error = geterror(bp);
		if (error) {
			goto rwexitout;
		}
		bufaddr = bp->b_un.b_addr;
	}
	amount = MIN(ip->i_size - offset,
	    MIN(bufremain, uiop->uio_resid));
	error = uiomove(bufaddr + bufoff, amount, UIO_READ, uiop);
	if (error)
		goto rwexitout;
	/* more? */
	if (uiop->uio_resid && ((offset + amount) < ip->i_size)) {
		offset += amount;
		bufremain = EXT2_BLOCK_SIZE(fsp);
		bufoff = 0;
		if (bp) {
			brelse(bp);
			bp = NULL;
		}
		blkno++;
		goto nextblk;
	}
rwexitout:
	if (bp) {
		if (error)
			bp->b_flags |= B_STALE | B_AGE;
		brelse(bp);
		bp = NULL;
	}
	rw_exit(&ip->i_rwlock);
	/*
	mutex_exit(&fsp->e_lock);
	 */

	if (holebuf)
		kmem_free(holebuf, EXT2_BLOCK_SIZE(fsp));
	return (error);
}

static int
ext2fs_write(struct vnode *vp, struct uio *uiop, int ioflag, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_write\n");
	return (ENOTSUP);
}

static int
ext2fs_getattr(struct vnode *vp, struct vattr *vap, int flags, struct cred *cr)
{
	struct inode *ip = VTOI(vp);
	struct ext2fs *fsp = ip->i_fs;
	int error;

	if (vap->va_mask == AT_SIZE) {
		/*
		 * for performance, if only the size is requested don't bother
		 * with anything else.
		 */
		vap->va_size = ip->i_size;
		return (0);
	}

	mutex_enter(&ip->i_tlock);
	/*
	 * Return all the attributes.  This should be refined so
	 * that it only returns what's asked for.
	 */

	/*
	 * Copy from inode table.
	 */
	vap->va_type = vp->v_type;
	vap->va_mode = ip->i_mode & MODEMASK;
	vap->va_uid = ip->i_uid;
	vap->va_gid = ip->i_gid;
	vap->va_fsid = ip->i_dev;
	vap->va_nodeid = (ino64_t)ip->i_number;
	vap->va_nlink = ip->i_links_count;
	vap->va_size = ip->i_size;
	vap->va_vcode = 0;
	if (vp->v_type == VCHR || vp->v_type == VBLK)
		vap->va_rdev = ip->i_dev;
	else
		vap->va_rdev = 0;	/* not a b/c spec. */
	vap->va_atime.tv_sec = (time_t)ip->i_atime;
	vap->va_atime.tv_nsec = 0;
	vap->va_mtime.tv_sec = (time_t)ip->i_mtime;
	vap->va_mtime.tv_nsec = 0;
	vap->va_ctime.tv_sec = (time_t)ip->i_ctime;
	vap->va_ctime.tv_nsec = 0;

	switch (ip->i_mode & IFMT) {

	case IFBLK:
		vap->va_blksize = MAXBSIZE;		/* was BLKDEV_IOSIZE */
		break;

	case IFCHR:
		vap->va_blksize = MAXBSIZE;
		break;

	default:
		vap->va_blksize = EXT2_BLOCK_SIZE(fsp);
		break;
	}
	vap->va_nblocks = (fsblkcnt64_t)ip->i_blocks;
	mutex_exit(&ip->i_tlock);
	error = 0;

out:
	return (error);
}

static int
ext2fs_setattr(struct vnode *vp, struct vattr *vap, int flags, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_setattr\n");
	return (ENOTSUP);
}

/*
 * Check mode permission on inode.  Mode is READ, WRITE or EXEC.
 * In the case of WRITE, the read-only status of the file system
 * is checked.  The mode is shifted to select the owner/group/other
 * fields.  The super user is granted all permissions except
 * writing to read-only file systems.
 */
static int
ext2fs_access(struct vnode *vp, int mode, int flags, struct cred *cr)
{
	struct inode *ip = VTOI(vp);
	int	error;

	if (ip->i_fs == NULL)
		return (EIO);
	error = ext2fs_iaccess(ip, mode, cr);
	return (error);
}

static int
ext2fs_lookup(struct vnode *dvp, char *nm, struct vnode **vpp,
    struct pathname *pnp, int flags, struct vnode *rdir, struct cred *cr)
{
	int	error = 0;
	struct inode *dp;
	struct inode *ip;
	struct ext2_dir_entry *ep;
	uint_t	offset;
	uint_t	entryoffsetinblock;
	uint_t	blkno;
	struct buf *bp = NULL;
	size_t	namelen;
	int	i;
	struct vnode *vp;
	uint32_t	ep_inode;
	uint16_t	ep_rec_len;
	uint16_t	ep_name_len;
	struct ext2fs	*fsp;

	dp = VTOI(dvp);
	fsp = dp->i_fs;

	/*
	 * Check accessibility of directory.
	 */
	if ((dp->i_mode & IFMT) != IFDIR)
		return (ENOTDIR);
	error = ext2fs_iaccess(dp, IEXEC, cr);
	if (error)
		return (error);
	/*
	 * Null component name is a synonym for directory being searched.
	 */
	if (*nm == '\0') {
		VN_HOLD(dvp);
		*vpp = dvp;
		return (0);
	}
	offset = 0;
	namelen = strlen(nm);
	mutex_enter(&fsp->e_lock);
	rw_enter(&dp->i_rwlock, RW_READER);
nextblk:
	entryoffsetinblock = 0;

	/* read in the next chunk */
	blkno = offset >> EXT2_BLOCK_SIZE_BITS(fsp);
	bp = ext2fs_bread(dp, blkno, 0, &error);
	if (bp == NULL) {
		goto rwexitout;
	}
	error = geterror(bp);
	if (error) {
		goto rwexitout;
	}

	while (entryoffsetinblock < EXT2_BLOCK_SIZE(fsp)) {
		ep = (struct ext2_dir_entry *)(bp->b_un.b_addr +
		    entryoffsetinblock);
		if ((ep->inode == 0) && (ep->rec_len == 0) &&
		    (ep->name_len == 0)) {
		    cmn_err(CE_WARN, "ext2fs_lookup: bad dir, inumber = %d\n",
			dp->i_number);
			error = ENXIO;
			goto rwexitout;
		}
		ep_inode = ext2fs_swap32(fsp, ep->inode);
		ep_rec_len = ext2fs_swap16(fsp, ep->rec_len);
		ep_name_len = ext2fs_swap16(fsp, ep->name_len);
		/*
		 * If the current directory entry is mangled, then skip
		 * to the next entry.  It would be nice to set the FSBAD
		 * flag in the super-block so that a fsck is forced on
		 * next reboot, but locking is a problem.
		 */
		if (ep_rec_len & (EXT2_DIR_PAD - 1)) {
			i = EXT2_DIR_PAD -
			    (entryoffsetinblock & (EXT2_DIR_PAD - 1));
			offset += i;
			entryoffsetinblock += i;
			continue;
		}
		/* check for a name match */
		if ((ep_inode != 0) && (ep_name_len == namelen) &&
		    *nm == *ep->name &&	/* fast chk 1st chr */
		    (bcmp(nm, ep->name, namelen) == 0)) {
			if (dp->i_number == ep_inode) {
				VN_HOLD(dvp);	/* want ourself, "." */
				*vpp = dvp;
			} else {
				error = ext2fs_iget(fsp->e_vfs, ep_inode,
				    &ip, cr);
				if (error)
					goto rwexitout;
				vp = ITOV(ip);
				*vpp = vp;
			}
			error = 0;
			goto rwexitout;
		}
		offset += ep_rec_len;
		entryoffsetinblock += ep_rec_len;
	}
	/* release the chunk */
	brelse(bp);
	bp = NULL;

	if (offset < dp->i_size)
		goto nextblk;

	error = ENOENT;

rwexitout:
	rw_exit(&dp->i_rwlock);
	mutex_exit(&fsp->e_lock);

out:
	if (bp) {
		if (error)
			bp->b_flags |= B_STALE | B_AGE;
		brelse(bp);
		bp = NULL;
	}
	return (error);
}

static int
ext2fs_create(struct vnode *dvp, char *name, struct vattr *vap, enum vcexcl excl,
    int mode, struct vnode **vpp, struct cred *cr, int flag)
{
	cmn_err(CE_CONT, "ext2fs_create\n");
	return (ENOTSUP);
}

static int
ext2fs_remove(struct vnode *vp, char *nm, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_remove\n");
	return (ENOTSUP);
}

static int
ext2fs_link(struct vnode *tdvp, struct vnode *svp, char *tnm, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_link\n");
	return (ENOTSUP);
}

static int
ext2fs_rename(struct vnode *sdvp, char *snm, struct vnode *tdvp, char *tnm,
    struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_rename\n");
	return (ENOTSUP);
}

static int
ext2fs_mkdir(struct vnode *dvp, char *dirname, struct vattr *vap,
    struct vnode **vpp, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_mkdir\n");
	return (ENOTSUP);
}

static int
ext2fs_rmdir(struct vnode *vp, char *nm, struct vnode *cdir, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_rmdir\n");
	return (ENOTSUP);
}

static int
ext2fs_readdir(struct vnode *vp, struct uio *uiop, struct cred *cr, int *eofp)
{
	struct ext2fs *fsp;
	struct inode *ip;
	int	error = 0;
	struct iovec *iovp;
	struct ext2_dir_entry *idp;
	struct dirent64 *odp;
	int	incount = 0;
	int	outcount = 0;
	size_t	bytes_wanted, total_bytes_wanted;
	caddr_t	outbuf = NULL;
	size_t	bufsize;
	uint_t	offset, offsetin;
	uint_t	blkno;
	struct buf *bp = NULL;
	uint32_t	idp_inode;
	uint16_t	idp_rec_len;
	uint16_t	idp_name_len;


	ip = VTOI(vp);

	if (uiop->uio_loffset >= MAXOFF32_T) {
		if (eofp)
			*eofp = 1;
		return (0);
	}
	fsp = ip->i_fs;
	mutex_enter(&fsp->e_lock);
	mutex_enter(&ip->i_tlock);
	iovp = uiop->uio_iov;
	total_bytes_wanted = iovp->iov_len;

	/* Large Files: directory files should not be "large" */
	ASSERT(ip->i_size <= MAXOFF32_T);

	/* Force offset to be valid (to guard against bogus lseek() values) */
	offset = (uint_t)uiop->uio_offset & ~(EXT2_DIR_PAD - 1);

	/* Quit if at end of file or link count of zero (posix) */
	if (offset >= (uint_t)ip->i_size || ip->i_links_count <= 0) {
		if (eofp)
			*eofp = 1;
		error = 0;
		goto out;
	}

	/*
	 * Get space to change directory entries into fs independent format.
	 * Do fast alloc for the most commonly used-request size (filesystem
	 * block size).
	 */
	if (uiop->uio_segflg != UIO_SYSSPACE || uiop->uio_iovcnt != 1) {
		/*
		 * The user level readdir() routines actually make
		 * getdent requests in units of 1048 instead of
		 * MAXBSIZE.  This seems a performance loss...
		 */
		bufsize = total_bytes_wanted;
		outbuf = kmem_alloc(bufsize, KM_SLEEP);
		odp = (struct dirent64 *)outbuf;
	} else {
		bufsize = total_bytes_wanted;
		odp = (struct dirent64 *)iovp->iov_base;
	}

nextblk:
	bytes_wanted = total_bytes_wanted;
	/* Truncate request to file size */
	if ((offset + bytes_wanted) > (int)ip->i_size)
		bytes_wanted = (int)ip->i_size;

	/* read in the next chunk */
	blkno = offset >> EXT2_BLOCK_SIZE_BITS(ip->i_fs);
	bp = ext2fs_bread(ip, blkno, 0, &error);
	if (bp == NULL) {
		error = EIO;
		goto out;
	}
	error = geterror(bp);
	if (error)
		goto out;
	incount = 0;
	idp = (struct ext2_dir_entry *)(bp->b_un.b_addr +
	    (offset - blkno * EXT2_BLOCK_SIZE(ip->i_fs)));
	if ((idp->inode == 0) && (idp->rec_len == 0) &&
	    (idp->name_len == 0)) {
	    cmn_err(CE_WARN, "ext2fs_readdir: bad dir, inumber = %d\n",
		ip->i_number);
		error = ENXIO;
		goto out;
	}

	/* Transform to file-system independent format */
	while (incount < bytes_wanted) {
		/*
		 * If the current directory entry is mangled, then skip
		 * to the next block.  It would be nice to set the FSBAD
		 * flag in the super-block so that a fsck is forced on
		 * next reboot, but locking is a problem.
		 */
		idp_inode = ext2fs_swap32(fsp, idp->inode);
		idp_rec_len = ext2fs_swap16(fsp, idp->rec_len);
		idp_name_len = ext2fs_swap16(fsp, idp->name_len);
		if (idp_rec_len & (EXT2_DIR_PAD - 1)) {
			offset = (offset + EXT2_DIR_PAD) & ~(EXT2_DIR_PAD-1);
			break;
		}
		/* Skip to requested offset and skip empty entries */
		if ((idp_inode != 0) && (offset >= (uint_t)uiop->uio_offset)) {
			ushort_t this_reclen = DIRENT64_RECLEN(idp_name_len);

			/* Buffer too small for any entries */
			if (!outcount && (this_reclen > bufsize)) {
				error = EINVAL;
				goto out;
			}
			/* If would overrun the buffer, quit */
			if ((outcount + this_reclen) > bufsize) {
				break;
			}
			/* Take this entry */
			odp->d_ino = (ino64_t)idp_inode;
			odp->d_reclen = (ushort_t)this_reclen;
			odp->d_off = (offset_t)(offset + idp_rec_len);
			(void) bcopy(idp->name, odp->d_name, idp_name_len);
			odp->d_name[idp_name_len] = '\0';
			outcount += odp->d_reclen;
			odp = (struct dirent64 *)((intptr_t)odp + odp->d_reclen);
			ASSERT(outcount <= bufsize);
		}
		if (idp_rec_len) {
			incount += idp_rec_len;
			offset += idp_rec_len;
			idp = (struct ext2_dir_entry *)((intptr_t)idp +
			    idp_rec_len);
		} else {
			offset = (offset + EXT2_DIR_PAD) & ~(EXT2_DIR_PAD-1);
			break;
		}
		/* need new block? */
		if (blkno != (offset >> EXT2_BLOCK_SIZE_BITS(ip->i_fs)))
			break;
	}
	/* release the chunk */
	brelse(bp);
	bp = NULL;

	if (offset < (int)ip->i_size && !outcount)
		goto nextblk;

	/* Copy out the entry data */

	error = 0;
	if (uiop->uio_segflg == UIO_SYSSPACE && uiop->uio_iovcnt == 1) {
		iovp->iov_base += outcount;
		iovp->iov_len -= outcount;
		uiop->uio_resid -= outcount;
		uiop->uio_offset = offset;
	} else if ((error = uiomove(outbuf, (long)outcount, UIO_READ,
	    uiop)) == 0) {
		uiop->uio_offset = offset;
	}

out:
	mutex_exit(&ip->i_tlock);
	mutex_exit(&fsp->e_lock);
	if (outbuf)
		kmem_free(outbuf, bufsize);
	if (eofp && (error == 0))
		*eofp = (uiop->uio_offset >= (int)ip->i_size);
	if (bp) {
		if (error)
			bp->b_flags |= B_STALE | B_AGE;
		brelse(bp);
	}
	return (error);
}

static int
ext2fs_symlink(struct vnode *dvp, char *linkname, struct vattr *vap,
    char *target, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_symlink\n");
	return (ENOTSUP);
}

static int
ext2fs_readlink(struct vnode *vp, struct uio *uiop, struct cred *cr)
{
	int	error = 0;
	struct buf *bp = NULL;
	char *link;
	struct inode *ip;
	int	i;
	struct ext2fs *fsp;
	uint32_t	i_data[EXT2_N_BLOCKS];

	if (vp->v_type != VLNK) {
		return (EINVAL);
	}
	ip = VTOI(vp);
	if (!S_ISLNK(ip->i_mode)) {
		return (EINVAL);
	}
	fsp = ip->i_fs;
	mutex_enter(&fsp->e_lock);
	rw_enter(&ip->i_rwlock, RW_READER);
	if (ip->i_blocks) {
		bp = ext2fs_bread(ip, 0, 0, &error);
		if (bp == NULL) {
			goto rwexitout;
		}
		error = geterror(bp);
		if (error) {
			goto rwexitout;
		}
		link = bp->b_un.b_addr;
	} else {
		/*
		 * one problem here is that we swapped the i_data blocks
		 * on reading the inode, so if it's a fast symlink we need
		 * to swap them back.
		 */
		for (i = 0; i < EXT2_N_BLOCKS; i++) {
			i_data[i] = ext2fs_swap32(fsp, ip->i_data[i]);
		}
		link = (char *)i_data;
	}
	error = uiomove(link, MIN(ip->i_size, uiop->uio_resid),
	    UIO_READ, uiop);
rwexitout:
	if (bp) {
		if (error)
			bp->b_flags |= B_STALE | B_AGE;
		brelse(bp);
	}
	rw_exit(&ip->i_rwlock);
	mutex_exit(&fsp->e_lock);

	return (error);
}

static int
ext2fs_fsync(struct vnode *vp, int syncflag, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_fsync\n");
	return (ENOTSUP);
}

static void
ext2fs_inactive(struct vnode *vp, struct cred *cr)
{
	struct ext2fs *fsp = VFSTOEXT2FS(vp->v_vfsp);
	kmutex_t	*ihm;
	int		hno;
	struct inode	*ip;

	mutex_enter(&fsp->e_lock);
	mutex_enter(&vp->v_lock);
	if (vp->v_count > 1) {
		vp->v_count--;
		mutex_exit(&vp->v_lock);
		mutex_exit(&fsp->e_lock);
		return;
	}

	/* remove it from the queue */
	ip = VTOI(vp);
	hno = INOHASH(vp->v_vfsp->vfs_dev, ip->i_number);
	ihm = &ih_lock[hno];
	mutex_enter(ihm);
	remque(ip);
	mutex_exit(ihm);

	mutex_exit(&vp->v_lock);

	if (vp->v_pages != NULL) {
		/*
		 * pvn_vplist_dirty will abort all old pages
		 */
		(void) pvn_vplist_dirty(vp, (u_offset_t)0, ext2fs_putapage,
		    B_INVAL, NULL);

	}
	kmem_cache_free(ext2fs_inode_cache, ip);
	mutex_exit(&fsp->e_lock);
}

static int
ext2fs_fid(struct vnode *vp, struct fid *fidp)
{
	struct efid *efid;
	struct inode *ip = VTOI(vp);

	if (ip->i_fs == NULL)
		return (EIO);
	if (fidp->fid_len < (sizeof (struct efid) - sizeof (ushort_t))) {
		fidp->fid_len = sizeof (struct efid) - sizeof (ushort_t);
		return (ENOSPC);
	}

	efid = (struct efid *)fidp;
	bzero(efid, sizeof (struct efid));
	efid->efid_len = sizeof (struct efid) - sizeof (ushort_t);
	efid->efid_ino = ip->i_number;
	efid->efid_version = ip->i_version;

	return (0);
}

static void
ext2fs_rwlock(struct vnode *vp, int write_lock)
{
	struct inode *ip = VTOI(vp);

	if (write_lock)
		rw_enter(&ip->i_rwlock, RW_WRITER);
	else
		rw_enter(&ip->i_rwlock, RW_READER);
}

static void
ext2fs_rwunlock(struct vnode *vp, int write_lock)
{
	struct inode *ip = VTOI(vp);

	rw_exit(&ip->i_rwlock);
}

static int
ext2fs_seek(struct vnode *vp, offset_t ooff, offset_t *noffp)
{
	return ((*noffp < 0 || *noffp > MAXOFFSET_T) ? EINVAL : 0);
}

static int
ext2fs_frlock(struct vnode *vp, int cmd, struct flock64 *bfp, int flag,
    offset_t offset, cred_t *cr)
{
	struct inode *ip = VTOI(vp);

	if (ip->i_fs == NULL)
		return (EIO);
	if (ip->i_mapcnt > 0)
		return (EAGAIN);
	return (fs_frlock(vp, cmd, bfp, flag, offset, cr));
}

static int
ext2fs_space(struct vnode *vp, int cmd, struct flock64 *bfp, int flag,
    offset_t offset, struct cred *cr)
{
	cmn_err(CE_CONT, "ext2fs_space\n");
	return (ENOTSUP);
}

static int
ext2fs_getapage(struct vnode *vp, u_offset_t off, size_t len, uint_t *protp,
    page_t *pl[], size_t plsz, struct seg *seg, caddr_t addr,
    enum seg_rw rw, struct cred *cr)
{
	struct inode *ip;
	struct ext2fs *fsp = VFSTOEXT2FS(vp->v_vfsp);
	int	error;
	struct vnode *devvp;
	page_t	*pp = NULL;
	page_t	*pagefound;

	ip = VTOI(vp);
	if (ip == NULL)
		return (EIO);
	devvp = fsp->e_devvp;

	/* no readaheads yet */
	if (pl == NULL)
		return (0);

	pl[0] = NULL;

reread:
	error = 0;

	if ((pagefound = page_exists(vp, off)) == NULL) {
		struct buf *bp;
		daddr_t	lbn, bn;
		u_offset_t io_off;
		size_t	io_len;
		size_t	pgoff;

		/*
		 * Need to really do disk IO to get the page
		 */
		lbn = off >> EXT2_BLOCK_SIZE_BITS(fsp);

		pp = pvn_read_kluster(vp, off, seg, addr, &io_off, &io_len,
		    off, EXT2_BLOCK_SIZE(fsp), 0);
		if (pp == NULL)
			goto reread;
		/* probably several I/Os */
		for (pgoff = 0; (pgoff < PAGESIZE) &&
		    ((off + pgoff) < ip->i_size);
		    lbn++, pgoff += EXT2_BLOCK_SIZE(fsp)) {
			/* get the actual block */
			bp = ext2_getblk(ip, lbn, 0, &error);
			if (bp == NULL) {
				if (error == EFBIG) {
					/* hole */
					error = 0;
					pagezero(pp->p_prev, pgoff,
					    EXT2_BLOCK_SIZE(fsp));
					continue;
				}
				error = EIO;
				goto out;
			}
			error = 0;
			bn = bp->b_blkno;
			brelse(bp);
			bp = pageio_setup(pp, EXT2_BLOCK_SIZE(fsp),
			    devvp, B_READ);
			bp->b_edev = devvp->v_rdev;
			bp->b_dev = cmpdev(devvp->v_rdev);
			bp->b_blkno = bn;
			bp->b_un.b_addr = (caddr_t)pgoff;
			(void) bdev_strategy(bp);
			error = biowait(bp);
			pageio_done(bp);
		}
		if (pgoff < PAGESIZE)
			pagezero(pp->p_prev, pgoff, PAGESIZE - pgoff);
		pvn_plist_init(pp, pl, plsz, off, io_len, rw);
	}
out:
	if (error) {
		if (pp != NULL)
			pvn_read_done(pp, B_ERROR);
		return (error);
	}

	if (pagefound) {
		/*
		 * Page exists in the cache, acquire the "shared"
		 * lock. If this fails, go back to reread.
		 */
		if ((pp = page_lookup(vp, off, SE_SHARED)) == NULL) {
			goto reread;
		}
		pl[0] = pp;
		pl[1] = NULL;
	}
	return (error);
}

static int
ext2fs_getpage(struct vnode *vp, offset_t off, size_t len, uint_t *protp,
    page_t *pl[], size_t plsz, struct seg *seg, caddr_t addr,
    enum seg_rw rw, struct cred *cr)
{
	struct ext2fs *fsp = VFSTOEXT2FS(vp->v_vfsp);
	struct inode *ip = VTOI(vp);
	int	error = 0;

	if (vp->v_flag & VNOMAP)
		return (ENOSYS);
	ASSERT(off <= MAXOFF32_T);

	/*
	mutex_enter(&fsp->e_lock);
	 */
	rw_enter(&ip->i_rwlock, RW_READER);
	if (protp != NULL)
		*protp = PROT_ALL;
	ASSERT((off & PAGEOFFSET) == 0);
	/* what if PAGESIZE > blocksize?, like on Ultra? */
	if (len <= PAGESIZE) {
		error = ext2fs_getapage(vp, off, len, protp, pl,
		    plsz, seg, addr, rw, cr);
	} else {
		error = pvn_getpages(ext2fs_getapage, vp, off,
		    len, protp, pl, plsz, seg, addr, rw, cr);
	}
	rw_exit(&ip->i_rwlock);
	/*
	mutex_exit(&fsp->e_lock);
	 */

	return (error);
}

static int
ext2fs_putapage(struct vnode *vp, page_t *pp, u_offset_t *offp, size_t *lenp,
    int flags, struct cred *cr)
{
	return (ENOTSUP);
}

static int
ext2fs_putpage(struct vnode *vp, offset_t off, size_t len, int flags,
    struct cred *cr)
{
	return (ENOTSUP);
}

static int
ext2fs_map(struct vnode *vp, offset_t off, struct as *as, caddr_t *addrp,
    size_t len, uchar_t prot, uchar_t maxprot, uint_t flags, struct cred *cr)
{
	struct segvn_crargs vn_a;
	int	error;

	if (vp->v_flag & VNOMAP)
		return (ENOSYS);
	if (((long)off < 0) || ((long)(off + len) < 0))
		return (EINVAL);
	if (off > MAXOFF32_T)
		return (EFBIG);

	as_rangelock(as);
	if ((flags & MAP_FIXED) == 0) {
		map_addr(addrp, len, off, 1, flags);
		if (*addrp == NULL) {
			as_rangeunlock(as);
			return (ENOMEM);
		}
	} else {
		/*
		 * User specified address - blow away any previous mappings
		 */
		(void) as_unmap(as, *addrp, len);
	}
	vn_a.vp = vp;
	vn_a.offset = off;
	vn_a.type = flags & MAP_TYPE;
	vn_a.prot = prot;
	vn_a.maxprot = maxprot;
	vn_a.flags = flags & ~MAP_TYPE;
	vn_a.cred = cr;
	vn_a.amp = NULL;

	error = as_map(as, *addrp, len, segvn_create, &vn_a);
	as_rangeunlock(as);
	return(error);
}

static int
ext2fs_addmap(struct vnode *vp, offset_t off, struct as *as, caddr_t addr,
    size_t len, uchar_t prot, uchar_t maxprot, uint_t flags, struct cred *cr)
{
	struct inode *ip = VTOI(vp);

	if (vp->v_flag & VNOMAP) {
		return (ENOSYS);
	}

	mutex_enter(&ip->i_tlock);
	ip->i_mapcnt += btopr(len);
	mutex_exit(&ip->i_tlock);
	return (0);
}

static int
ext2fs_delmap(struct vnode *vp, offset_t off, struct as *as, caddr_t addr,
    size_t len, uint_t prot, uint_t maxprot, uint_t flags, struct cred *cr)
{
	struct inode *ip = VTOI(vp);

	if (vp->v_flag & VNOMAP) {
		return (ENOSYS);
	}

	mutex_enter(&ip->i_tlock);
	ip->i_mapcnt -= btopr(len);     /* Count released mappings */
	ASSERT(ip->i_mapcnt >= 0);
	mutex_exit(&ip->i_tlock);
	return (0);
}

static int
ext2fs_poll(vnode_t *vp, short ev, int any, short *revp, struct pollhead **phpp)
{
	cmn_err(CE_CONT, "ext2fs_poll\n");
	return (ENOTSUP);
}

static int
ext2fs_l_pathconf(struct vnode *vp, int cmd, ulong_t *valp, struct cred *cr)
{
	return (fs_pathconf(vp, cmd, valp, cr));
}

static int
ext2fs_dumpctl(vnode_t *vp, int free)
{
	cmn_err(CE_CONT, "ext2fs_dumpctl\n");
	return (ENOTSUP);
}

static int
ext2fs_dump(vnode_t *vp, caddr_t addr, int ldbn, int dblks)
{
	cmn_err(CE_CONT, "ext2fs_dump\n");
	return (ENOTSUP);
}
